﻿namespace HeuristicLab.Problems.ProgramSynthesis.Push.Interpreter {
  using System;
  using HeuristicLab.Core;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Configuration;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Data.Pool;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Expressions;

  public class PushInterpreterPool {
    private readonly ObjectPool<PooledPushInterpreter> pool;

    public readonly ManagedPoolProvider<PushProgram> PushProgramPoolProvider;
    public readonly ManagedPoolProvider<PooledList<Expression>> ExpressionListPoolProvider;

    public PushInterpreterPool(IReadOnlyPushConfiguration config = null)
      : this(Environment.ProcessorCount * 4, 1024, null, config) {
    }

    public PushInterpreterPool(int size, int poolPartitionSize, int? maxPartitionCount = null, IReadOnlyPushConfiguration config = null) {
      PushConfiguration = config ?? new PushConfiguration();

      PushProgramPoolProvider = new ManagedPoolProvider<PushProgram>(poolPartitionSize, () => new PushProgram(), maxPartitionCount);
      ExpressionListPoolProvider = new ManagedPoolProvider<PooledList<Expression>>(poolPartitionSize, () => new PooledList<Expression>(), maxPartitionCount);

      pool = new ObjectPool<PooledPushInterpreter>(() => {
        var poolContainer = new InterpreterPoolContainer(PushProgramPoolProvider, ExpressionListPoolProvider);
        return new PooledPushInterpreter(this, PushConfiguration, poolContainer);
      }, size);
    }

    public IReadOnlyPushConfiguration PushConfiguration { get; private set; }

    public PooledPushInterpreter Create(IRandom random = null) {
      var interpreter = pool.Allocate();
      interpreter.Reset(random);

      return interpreter;
    }

    public void Free(PooledPushInterpreter interpreter) {
      interpreter.ClearStacks();
      pool.Free(interpreter);
    }

    public void Clear() {
      PushProgramPoolProvider.Clear();
      ExpressionListPoolProvider.Clear();
      pool.Clear();
    }
  }
}
